without warning
without trace
include wildcard.e
-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- MATH
--/topic Math Routines
--/info
-- Miscellaneous mathematical and low-level routines

--/topic Math Routines
--/func get_bits( atom b32 )
--/desc Does the reverse of /or_all() in that it gets all the bit values from an atom.
--/ret Sequence: The non-zero bits set in /i b32.
--
-- Typically used to extract which bits have been set on in a flag.
--
-- Example:
--
-- /code        
--      -- combine flags
--      integer flags
--      sequence codes
--      codes = get_bits( origflag )
--      if find(WS_EX_CLIENTEDGE, codes) then
--          setText(statusarea, "Client Edge has been specified")
--      end if
-- /endcode
constant kBitPosn = {
    #00000001 , #00000002, #00000004, #00000008,
    #00000010 , #00000020, #00000040, #00000080,
    #00000100 , #00000200, #00000400, #00000800,
    #00001000 , #00002000, #00004000, #00008000,
    #00010000 , #00020000, #00040000, #00080000,
    #00100000 , #00200000, #00400000, #00800000,
    #01000000 , #02000000, #04000000, #08000000,
    #10000000 , #20000000, #40000000, #80000000
                     }
global function get_bits(atom b32)
sequence lResult

    lResult = {}
    for i = 1 to length(kBitPosn) do
        if and_bits(b32, kBitPosn[i]) then
            lResult &= kBitPosn[i]
        end if
    end for

    if length(lResult) = 0 then
        lResult = {0}
    end if
    
    return lResult
end function


-----------------------------------------------------------------------------

--/topic Math Routines
--/func signed_word( atom a )
--/desc Converts /i a into a signed 16-bit integer.
--/ret INTEGER: Signed 16 bits of a 32 bit word.
-- Typically used to decode Win32 message values when several values 
-- are packed into a single number. 
-- 
-- Example: 
--
-- /code
--      -- extract the low portion from lParam
--      integer x
--      x = signed_word( lParam )
-- /endcode

global function signed_word( atom a )
    -- return low word
    integer x
    x = and_bits( a, #FFFF )
    if x > #7FFF then
        x -= #10000
    end if
    return x
end function


--/topic Math Routines
--/func unpack_sWord( atom a )
--/desc Returns signed high and signed low portions of a 32-bit word.
--/ret SEQUENCE: {High signed 16 bits,Low signed 16 bits}
-- Typically used to decode Win32 message values when several values 
-- are packed into a single number.
--
-- Example:
--
-- /code         
--      -- extract the signed words from lParam
--      sequence hilo
--      hilo = upack_sWord( lParam )
-- /endcode

global function unpack_sWord( atom a )
    return {signed_word(floor(a / #10000 )), signed_word(a)}
end function


-----------------------------------------------------------------------------
--/topic Math Routines
--/func pack_word( integer low, integer high )
--/desc Packs values into word.
--/ret ATOM: 32 bit word with low value in low 16 bits, high value in high 16 bits.
-- Typically used to encode a message parameter into a 32 bit word.
--
-- Example:
--
-- /code         
--      -- pack min and max into parameter
--      integer y
--      lParam = pack_word( min, max )
-- /endcode

global function pack_word( integer low, integer high )
    -- return packed 32 bit value
    
    return ( and_bits( high, #FFFF ) * #10000 ) + and_bits( low,  #FFFF )
    
end function


--/topic Math Routines
--/func lo_word( atom pData)
--/desc returns the low-16 bits of /i pData
--/ret INTEGER: Bits 15-0 of the parameter
--
-- Example:
-- /code         
--      integer y
--      y = lo_word( bigval )
-- /endcode
           
global function lo_word( atom pData)
    return and_bits(pData, #FFFF)
end function         

--/topic Math Routines
--/func hi_word( atom pData)
--/desc returns the high 16 bits of /i pData
--/ret INTEGER: Bits 31-16 of the parameter as a 16 bit value.
--
-- Example:
-- /code         
--      integer y
--      y = hi_word( bigval )
-- /endcode

global function hi_word( atom pData)
    return and_bits(and_bits(pData, #FFFF0000) / #10000, #FFFF)
end function         

-----------------------------------------------------------------------------
--/topic Math Routines
--/func abs( object a )
--/desc Absolute value.
--/ret Absolute value of the atom or sequence.
--
--Example
--/code
--       sequence s
--       atom a
--
--       a = abs(-1)   -- = 1
--       s = abs({-2,-1,{-1,0,1},1,2}) -- = {2,1,{1,0,1},1,2}
--/endcode

global function abs( object a )
    if atom(a) then
        if a < 0 then
            return -a
        else
            return a
        end if
    else
        for i = 1 to length(a) do
            a[i] = abs(a[i])
        end for

        return a
    end if
    
end function


----------------------------------------------------------------------------
--/topic Math Routines
--/func shortInt( atom a )
--/desc Converts a number into a 16-bit signed integer
--/ret INTEGER: The value of /i a as a short int.
-- Typically used to decode Win32 message values
--
-- Example:
--
-- /code         
--      -- Mouse Position
--      sequence hilo, x, y
--      hilo = upack_word( lParam )
--      x = shortInt(hilo[2])
--      y = shortInt(hilo[1])
-- /endcode
global function shortInt( atom i )

    -- converts numbers ( 4 bytes #0000 to #FFFF)
    -- to signed short ints (2 bytes -32768 to 32767 )

    -- Force the use of only the rightmost 2 bytes.
    i = and_bits(i, #FFFF)

    if i >= 0 and i <= #7FFF then
        return i
    else
        return i - #10000
    end if
    
end function



--/topic Math Routines
--/func TextToNumber( sequence text )
--/ret Atom: The number represented by the text.
--/desc This converts the text into a number.
-- If the text contains invalid characters, zero is returned.
--
-- /b "Note 1:" You can supply Hexadecimal values if the value is preceded by
-- a '#' character, Octal values if the value is preceded by a '@' character,
-- and Binary values if the value is preceded by a '!' character. With
-- hexadecimal values, the case of the digits 'A' - 'F' is not important. Also,
-- any period character embedded in the number is used with the correct base.
--
-- /b "Note 2:" Any underscore characters, '_', that are embedded in the text 
-- number are ignored. These can be used to help visual clarity for long numbers.
--
--/b "Note 3:" You can supply a leading or trailing, minus or plus sign.
--
--/b "Note 4:" You can supply trailing percentage sign(s). Each one present causes
-- the resulting value to be divided by 100.
--
-- This function can optionally return information about invalid numbers. If
-- /i text has the form of {sequence, integer} then if the integer is nonzero,
-- a sequence is returned. The first element is the value converted, and the
-- second is the position in the text where conversion stopped. If no errors
-- were found then this is zero.
--
--/code
--     sequence rc
--     atom   val
--     rc = TextToNumber({"12.34a", 1})
--     --  rc ---> {12.34, 6} -- Error at position 6
--     rc = TextToNumber({"12.34", 1})
--     --  rc ---> {12.34, 0} -- No errors.
--
--     val = TextToNumber("12.34a")
--     --  val ---> 0
--
--      val = TextToNumber("#f80c") --> 63500
--      val = TextToNumber("#f80c.7aa") --> 63500.47900390625
--      val = TextToNumber("@1703") --> 963
--      val = TextToNumber("!101101") --> 45                            
--      val = TextToNumber("12_583_891") --> 12583891
--      val = TextToNumber("12_583_891%") --> 125838.91
--      val = TextToNumber("12_583_891%%") --> 1258.3891
--
--/endcode                 
constant vDigits = ".0123456789ABCDEF"
global function TextToNumber( sequence text)
    -- get the numeric value of text
    integer dot,sign,tstart,tend, v, note, notify
    atom lhs, rhs, lh, rh 
    integer base, pc
    atom value

    dot = 0
    lh = 0
    lhs = 0
    rh = 0
    rhs = 1
    sign = 0
    note = 0
    base = 10       
    pc = 1

    if  length(text) = 2  and
        sequence(text[1]) and
        integer(text[2])
    then
        notify  = text[2]
        text = text[1]
    else
        notify = 0
    end if

    -- convert the value of the text    
    text = upper(text)
    tstart = 1
    tend = length(text)
    -- Ignore leading whitespace
    while tstart <= tend do
        if equal(text[tstart], '-') and sign = 0 then
            sign = -1
        elsif equal (text[tstart],'+') and sign = 0 then
            sign = 1
        elsif equal (text[tstart],'#') then
            base = 16
        elsif equal (text[tstart],'@') then
            base = 8
        elsif equal (text[tstart],'!') then
            base = 2
        elsif find(text[tstart], {'\t', ' '}) = 0 then
            exit
        end if
        tstart += 1
    end while   
    
    -- Ignore trailing whitespace
    while tstart <= tend do
        if equal(text[tend], '-') and sign = 0 then
            sign = -1
        elsif equal(text[tend],'+') and sign = 0 then
            sign = 1
        elsif equal(text[tend],'%') then
            pc *= 100
        elsif find(text[tend], {'\t', ' '}) = 0 then
            exit
        end if
        tend -= 1
    end while
    
    -- Set the default sign.
    if sign = 0 then
        sign = 1
    end if
    
    for i = tstart to tend do
        if lhs > 0 and text[i] = '_' then
            -- ignore an embedded underscore.
        else
            v =  find(text[i], vDigits)
            -- Invalid char so force a zero return.
            if v = 1 then -- A dot found.
                if dot = 0 then
                    dot = 1
                else
                    note = i
                    if notify = 0 then
                        sign = 0
                    end if
                    exit
                end if
            else       
                v -= 1
                if v < 0 or v > base then
                    -- Illegal char found.
                    note = i
                    if notify = 0 then
                        sign = 0
                    end if
                    exit
                else
                    if dot = 1 then
                        rhs *= base
                        rh = (rh * base) + v - 1
                    else
                        lhs += 1
                        lh = (lh * base) + v - 1
                    end if
                end if
             end if
        end if
        
        -- I got to the end without error!
        if i = tend then
            note = 0
        end if
         
    end for
                    
    value = ((lh + (rh / rhs)) * sign) / pc
    if notify = 0 then
        return value
    else
        return {value, note}
    end if
end function

-----------------------------------------------------------------------------
--/topic Math Routines
--/func unpack_word( atom a )
--/desc Returns high and low portions of a 32-bit word.
--/ret SEQUENCE: {High 16 bits,Low 16 bits}
-- Typically used to decode Win32 message values when several values 
-- are packed into a single number.
--
-- Example:
--
-- /code         
--      -- extract the short words from lParam
--      sequence hilo
--      hilo = upack_word( lParam )
-- /endcode

function unpack_word( atom a )
    return {floor(a / #10000 ), and_bits(a, #FFFF)}
end function

--/topic Math Routines
--/func or_all( object pData )
--/desc Calculates a binary OR against each element in /pData
--/ret ATOM: The OR'd value
--
--example:
--/code
--      atom flags
--      flags = or_all({WS_CHILD, WS_VISIBLE, WS_BORDER})
--/endcode

global function or_all( object pData )
    atom lResult
    
    if atom(pData) then
        return pData
    end if
    
    if length(pData) = 0 then
        return 0
    end if
    
    if length(pData) = 1 then
        return or_all(pData[1])
    end if
    
    lResult = or_bits(pData[1], pData[2])
    for i = 3 to length(pData) do
        lResult = or_bits(lResult, pData[i])
    end for
    
    return lResult
end function    
        
global function clear_bits(atom a, atom mask)
    return and_bits(a, not_bits(mask))
end function