-- Library: tk_trim.e
-- contains...
--   trim_left_chars(s, s)
--   trim_right_chars(s, s)
--   trim_left(s)
--   trim_right(s)
--   trim(s)
without trace

--/topic String Handling
--/info
--This set of utility functions remove specified characters from sequences

--/topic String Handling
--/func trim_left_chars(sequence pSource, sequence pChars)
--/desc Removes the leftmost set of characters
--/ret The original sequence with the matching characters removed.
--/i pSource is the sequence from which to remove characters. /n
--/i pChars is the sequence containing which characters can be removed. /n
--Example:
--/code
--       sequence result
--       result = trim_left_chars("321abc9def123", "1234567890")
--       ? result  -- Should display "abc9def123"
--/endcode

------------------------------------------------------
global function trim_left_chars(sequence pSource, sequence pChars)
------------------------------------------------------
integer lStart, lEnd

    lStart = 1
    lEnd   = length(pSource)

    while lStart <= lEnd do
        if not find(pSource[lStart], pChars) then
            return pSource[lStart..lEnd]
        end if
        lStart += 1
    end while
    
    return ""
end function

--/topic String Handling
--/func trim_right_chars(sequence pSource, sequence pChars)
--/desc Removes the rightmost set of characters
--/ret The original sequence with the matching characters removed.
--/i pSource is the sequence from which to remove characters. /n
--/i pChars is the sequence containing which characters can be removed. /n
--Example:
--/code
--       sequence result
--       result = trim_left_chars("321abc9def123", "1234567890")
--       ? result  -- Should display "321abc9def"
--/endcode

------------------------------------------------------
global function trim_right_chars(sequence pSource, sequence pChars)
------------------------------------------------------
integer  lEnd

    lEnd   = length(pSource)
   
    while lEnd do
        if not find(pSource[lEnd], pChars) then
            return pSource[1..lEnd]
        end if
        lEnd -= 1
    end while

    return ""
end function

--/topic String Handling
--/func trim_chars(sequence pSource, sequence pChars)
--/desc Removes any /i pChars from both ends of /i pSource
--/ret The original sequence with the matching characters removed.
--/i pSource is the sequence from which to remove whitespace. /n
--Example:
--/code
--       sequence result
--       result = trim("123abc9def321", "3459")
--       ? result  -- Should display "12abcdef21"
--/endcode

------------------------------------------------------
global function trim_chars(sequence pSource, sequence pChars)
------------------------------------------------------
    return trim_left_chars(trim_right_chars(pSource, pChars), pChars)

end function

--/topic String Handling
--/func trim_all_chars(sequence pSource, sequence pChars)
--/desc Removes all the matching characters.
--/ret The original sequence with the matching characters removed.
--/i pSource is the sequence from which to remove characters. /n
--/i pChars is the sequence containing which characters can be removed. /n
--Example:
--/code
--       sequence result
--       result = trim_all_chars("321abc9def123", "1234567890")
--       ? result  -- Should display "abcdef"
--/endcode

------------------------------------------------------
global function trim_all_chars(sequence pSource, sequence pChars)
------------------------------------------------------
sequence lResult
integer  lLast

    lLast = 0
    lResult = repeat(0, length(pSource))

    for i = 1 to length(pSource) do
        if find(pSource[i], pChars) = 0 then
            lLast += 1
            lResult[lLast] = pSource[i]
        end if
    end for
    
    return lResult[1 .. lLast]
end function


--/topic String Handling
--/func compress_chars(sequence pSource, sequence pChars, integer pRep)
--/desc Replaces embedded sequences of pChars with /i pRep
--/ret The original sequence with the matching characters replaced.
--/i pSource is the sequence from which to replace characters. /n
--/i pChars is the sequence containing which characters can be replaced. /n
--/i pRep is the replacement character.
--The /i pSource string is first trimmed of any /i pChars, then sets of one
-- or more /i pChars is replaced by a single /i pRep character.
--
--Example:
--/code
--  sequence result
--  result = compress_chars("\t\tif  abc  = def  then xyz()  \t", " \t", ',')
--  ? result  -- Should display "if,abc,=,def,then,xyz()"
--/endcode

global function compress_chars(sequence pSource, sequence pChars, integer pRep)
    integer low, last, cur, ix, cnt, token, reptoken

    -- No brainer: If no chars to replace then just send back what was sent.
    if length(pChars) = 0 then
        return pSource
    end if

    -- Trim the string of pChars.
    pSource = trim_chars(pSource, pChars)

    -- If nothing left, don't bother continuing.
    if length(pSource) = 0 then
        return pSource
    end if      

    -- Find a token character to use.
    token = -1
    while find(token, pSource & pRep & pChars) do
        token -= 1
    end while

    -- Find a replacement token character to use.
    reptoken = token - 1
    while find(reptoken, pSource & pRep) do
        reptoken -= 1
    end while

    -- Set all pChars to reptoken
    cnt = 0
    for i = 1 to length(pChars) do
        ix = find(pChars[i], pSource)
        while ix do
            pSource[ix] = reptoken
            ix = find(pChars[i], pSource)
            cnt += 1
        end while
    end for

    -- If there isn't any instances of pChars then return
    if cnt = 0 then
        return pSource
    end if
            
    -- Initialise
    last = length(pSource)
    cnt = 0
    low = find(reptoken, pSource)
    while low do
        pSource[low] = token -- Force a token
        cnt += 1
        low += 1
        
        cur = low
        while pSource[cur] = reptoken do
            cur += 1
        end while

        if low != cur then
            pSource[low .. last - cur + low ] = pSource[cur .. last]
            last -= (cur - low)
        end if
        
        low = find(reptoken, pSource)
        if low > last then
            exit
        end if
    end while

    if cnt != 0 then
        -- Make all tokens a pRep
        ix = find(token, pSource)
        while ix do
            pSource[ix] = pRep
            ix = find(token, pSource)
        end while
        return pSource[1 .. last]
    else
        return pSource
    end if

end function


----------------- Predefined Whitespace Routines --------------------

sequence whitespace
whitespace = " " & 9 & 10 & 11 & 12 & 13

--/topic String Handling
--/func set_whitespace(sequence pChars)
--/desc Defines what is "whitespace"
--/ret The current whitespace set of characters.
--/i pChars is the new set of whitespace characters. /n
--Initially, the set of whitespace characters is /n
-- Space, Tab, LineFeed, Carriage Return, Form Feed and Vertical Tab /n
-- These correspond to ASCII 32, 9, 10, 13, 12, 11
--
--/b<Note:> If /i pChars is an empty sequence, the current set will not be
-- changed. It will still be returned though.
--
--Example:
--/code
--       sequence oldWS, result
--       oldWS = set_whitespace(",.()[]{}")
--       result = trim_all("123 Smith St., Collingwood, 3728, (VIC).")
--       ? result -- Should display "123 Smith St Collingwood 3728 VIC"
--       oldWS = set_whitespace(oldWS)
--/endcode

------------------------------------------------------
global function set_whitespace_chars(sequence pChars)
------------------------------------------------------
    sequence oldws

    oldws = whitespace
    if length(pChars) > 0 then
        whitespace = pChars
    end if
    
    return oldws
end function

--/topic String Handling
--/func trim(sequence pSource)
--/desc Removes any whitespace chars from both ends of /i pSource
--/ret The original sequence with the matching characters removed.
--/i pSource is the sequence from which to remove whitespace. /n
--Example:
--/code
--       sequence result
--       result = trim("  abc def  ")
--       ? result  -- Should display "abc def"
--/endcode

------------------------------------------------------
global function trim(sequence pSource)
------------------------------------------------------
    
    return trim_chars(trim_right_chars(pSource, whitespace), whitespace)
    
end function

--/topic String Handling
--/func trim_left(sequence pSource)
--/desc Removes any whitespace chars from the left of /i pSource
--/ret The original sequence with the matching characters removed.
--/i pSource is the sequence from which to remove whitespace. /n
--Example:
--/code
--       sequence result
--       result = trim_left("  abc def  ")
--       ? result  -- Should display "abc def  "
--/endcode
------------------------------------------------------
global function trim_left(sequence pSource)
------------------------------------------------------
    return trim_left_chars(pSource, whitespace)
end function

--/topic String Handling
--/func trim_right(sequence pSource)
--/desc Removes any whitespace chars from the right of /i pSource
--/ret The original sequence with the matching characters removed.
--/i pSource is the sequence from which to remove whitespace. /n
--Example:
--/code
--       sequence result
--       result = trim_right("  abc def  ")
--       ? result  -- Should display "  abc def"
--/endcode

------------------------------------------------------
global function trim_right(sequence pSource)
------------------------------------------------------
    return trim_right_chars(pSource, whitespace)
end function

--/topic String Handling
--/func trim_all(sequence pSource)
--/desc Removes all whitespace chars from /i pSource
--/ret The original sequence with the matching characters removed.
--/i pSource is the sequence from which to remove whitespace. /n
--Example:
--/code
--       sequence result
--       result = trim_all("  abc def  ")
--       ? result  -- Should display "abcdef"
--/endcode

------------------------------------------------------
global function trim_all(sequence pSource)
------------------------------------------------------
    return trim_all_chars(pSource, whitespace)
end function

--/topic String Handling
--/func compress(sequence pSource, integer pRep)
--/desc Replaces embedded whitespace with /i pRep
--/ret The original sequence with the matching characters replaced.
--/i pSource is the sequence from which to replace characters. /n
--/i pRep is the delimiter character or string.
--The /i pSource string is first trimmed of any whitespace, then sets of one
-- or more whitespace characters are replaced by a single /i pRep character.
--
--Example:
--/code
--  sequence result
--  result = compress("\t\tif  abc  = def  then xyz()  \t", ' ')
--  ? result  -- Should display "if abc = def then xyz()"
--/endcode

------------------------------------------------------
global function compress(sequence pSource, integer pRep)
------------------------------------------------------
    return compress_chars(pSource, whitespace, pRep)
end function


--/topic String Handling
--/func split(sequence pSource, object pDelim)
--/desc Returns the undelimited substrings
--/ret SEQUENCE: A set of sequences, one per delimited substring in /i pSource
--/i pSource is the sequence from which to extract the substrings. /n
--/i pDelim is the delimiter character, delimiter string, or a set of
-- single character delimiters. /n
--
--/b "Note 1" - if /i pDelim is an atom and not equal to '{' then sections 
--of text enclosed in matching '{' and '}' are not scanned. /n
--/b "Note 2" - if /i pDelim is a single string inside a sequence, then
-- it is considered to be a set of delimiter characters and any one of
-- them can delimiter the /i pSource. The results in this case are that
-- each delimitered text is followed by the sequence containing the 
-- character that caused the delimitation. A final empty sequence means
-- that no delimiter was found to end the /i pSource text.
--
--Example:
--/code
--  sequence result
--  result = split("if abc = def then xyz()", ' ')
--  ? result  -- Should be {"if","abc","=","def","then","xyz()"}
--
--  -- Example of embedded delimiter...
--  result = split("event=Click,font={Courier,12}", ',')
--  ? result  -- Should be {"event=Click","font={Courier,12}"}
--
--  -- Example of delimiter set...
--  result = split("event=Click,font={Courier,12}", {"=,{}"})
--  ? result  -- Should be {"event","=","Click",",","font","=","","{","Courier",",","12","}"}
--
--/endcode
------------------------------------------------------
global function split(sequence pSource,  object pDelim)
------------------------------------------------------
    sequence lResult
    integer l,r,e
    integer lDepth
    
    lResult = {} 
    lDepth = 0
    l = 1
    r = 1
    if atom(pDelim) then
        -- Single char delimiter.
        while r <= length(pSource) do
            if lDepth = 0 then
                if pSource[r] = pDelim then
                    lResult = append(lResult, pSource[l..r-1])
                    l = r+1
                    r = l
                else
                    if pSource[r] = '{' then
                        lDepth += 1
                    end if
                    r += 1
                end if
            else
                if pSource[r] = '{' then
                    lDepth += 1
                elsif pSource[r] = '}' then
                    lDepth -= 1
                end if 
                r += 1
            end if
        end while
        lResult = append(lResult, pSource[l..r-1])
            
    elsif length(pDelim) > 0 then
        if atom(pDelim[1]) then 
            -- Delimiter is assumed to be a simple string.
            e = r+length(pDelim)-1
            while e <= length(pSource) do
                if equal(pSource[r..e], pDelim) then
                    lResult = append(lResult, pSource[l..r-1])
                    l = e+1
                    r = l
                    e = r+length(pDelim)-1
                else
                    r += 1
                    e += 1
                end if
            end while  
            
            lResult = append(lResult, pSource[l..length(pSource)])
        else
            -- In this case, the format is a string inside a sequence
            -- and the string is a list of single char delimiters.  
            pDelim = pDelim[1]
            while r <= length(pSource) do
                if find(pSource[r], pDelim) then
                    lResult = append(lResult, pSource[l..r-1])
                    lResult = append(lResult, {pSource[r]})
                    l = r+1
                    r = l
                else
                    r += 1
                end if
            end while
            lResult = append(lResult, pSource[l..r-1])
            lResult = append(lResult, {})
        end if                       
    else
        return pSource
    end if
  return lResult    
end function
                                

--/topic String Handling
--/func lookup(object pItem, sequence pSource, sequence pTarget)
--/desc Returns the corresponding element.
--/ret OBJECT: The corresponding element or 0/{} if not found.
--This searches /i pSource for /i pItem and if found, it returns the
-- corresponding element from /i pTarget. /n
--If /i pItem is not found in /i pSource, then what is returned depends
-- on a few things. /n
--If /i pTarget is longer than /i pSource then the last element in
-- /i pTarget is returned. This implements a way to return a default
-- value of your choosing. /n
--If /i pTarget is not longer than /i pSource then if the first element
-- of /i pTarget is an atom then zero is returned otherwise an empty
-- sequence is returned.
--
--Examples:
--
--/code
--       x = lookup('a', "cat", "dog") --> 'o'
--       x = lookup('d', "cat", "dogx") --> 'x'
--       x = lookup('d', "cat", "dog") --> 0
--       x = lookup("ape", {"ant","bear","cat"}, {"spider","seal","dog"}) --> ""
--       x = lookup("ant", {"ant","bear","cat"}, {"spider","seal","dog"}) --> "spider"
--/endcode
global function lookup(object pItem, sequence pSource, sequence pTarget)
    integer lPosn
    
    lPosn = find(pItem, pSource)
    if lPosn > 0 then   
        return pTarget[lPosn]
    elsif length(pTarget) > length(pSource) then
        return pTarget[length(pTarget)]
    elsif atom(pTarget[1]) then
        return 0
    else
        return ""
    end if

end function